@@ -4,11 +4,10 @@
  *	Contains the Unix-specific interpreter initialization functions.
  *
  * Copyright (c) 1995-1997 Sun Microsystems, Inc.
+ * Copyright (c) 1999 by Scriptics Corporation.
+ * All rights reserved.
  *
- * See the file "license.terms" for information on usage and redistribution
- * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclUnixInit.c,v 1.14 1999/04/21 22:00:29 stanton Exp $
+ * RCS: @(#) $Id: tclUnixInit.c,v 1.17 1999/07/22 21:50:56 redman Exp $
  */
 
 #include "tclInt.h"
@@ -46,6 +45,11 @@ static char defaultLibraryDir[sizeof(TCL_LIBRARY)+200] = TCL_LIBRARY;
 
 static char pkgPath[sizeof(TCL_PACKAGE_PATH)+200] = TCL_PACKAGE_PATH;
 
+/*
+ * The following table is used to map from Unix locale strings to
+ * encoding files.
+ */
+
 typedef struct LocaleTable {
     CONST char *lang;
     CONST char *encoding;
@@ -80,6 +84,10 @@ static CONST LocaleTable localeTable[] = {
     {"ko_KR.eucKR",     "euc-kr"},
     {"korean",          "euc-kr"},
 
+    {"ru",		"iso8859-5"},		
+    {"ru_RU",		"iso8859-5"},		
+    {"ru_SU",		"iso8859-5"},		
+
     {"zh",		"cp936"},
 
     {NULL, NULL}
@@ -259,17 +267,17 @@ CONST char *path;		/* Path to the executable in native
      * This code looks in the following directories:
      *
      *	<bindir>/../<installLib>
-     *		(e.g. /usr/local/bin/../lib/tcl8.1)
+     *		(e.g. /usr/local/bin/../lib/tcl8.2)
      *	<bindir>/../../<installLib>
-     *		(e.g. /usr/local/TclPro/solaris-sparc/bin/../../lib/tcl8.1)
+     *		(e.g. /usr/local/TclPro/solaris-sparc/bin/../../lib/tcl8.2)
      *	<bindir>/../library
-     *		(e.g. /usr/src/tcl8.1/unix/../library)
+     *		(e.g. /usr/src/tcl8.2/unix/../library)
      *	<bindir>/../../library
-     *		(e.g. /usr/src/tcl8.1/unix/solaris-sparc/../../library)
+     *		(e.g. /usr/src/tcl8.2/unix/solaris-sparc/../../library)
      *	<bindir>/../../<developLib>
-     *		(e.g. /usr/src/tcl8.1/unix/../../tcl8.1/library)
+     *		(e.g. /usr/src/tcl8.2/unix/../../tcl8.2/library)
      *	<bindir>/../../../<devlopLib>
-     *		(e.g. /usr/src/tcl8.1/unix/solaris-sparc/../../../tcl8.1/library)
+     *		(e.g. /usr/src/tcl8.2/unix/solaris-sparc/../../../tcl8.2/library)
      */
      
     if (path != NULL) {
@@ -361,15 +369,19 @@ TclpSetInitialEncodings()
     int i;
     Tcl_Obj *pathPtr;
     char *langEnv;
+    Tcl_DString ds;
 
     /*
-     * Determine the current encoding from the LC_TYPE or LANG environment
+     * Determine the current encoding from the LC_* or LANG environment
      * variables.  We previously used setlocale() to determine the locale,
      * but this does not work on some systems (e.g. Linux/i386 RH 5.0).
      */
 
-    langEnv = getenv("LC_CTYPE");
+    langEnv = getenv("LC_ALL");
 
+    if (langEnv == NULL || langEnv[0] == '\0') {
+	langEnv = getenv("LC_CTYPE");
+    }
     if (langEnv == NULL || langEnv[0] == '\0') {
 	langEnv = getenv("LANG");
     }
@@ -377,17 +389,73 @@ TclpSetInitialEncodings()
 	langEnv = NULL;
     }
 
-    encoding = "iso8859-1";
+    encoding = NULL;
     if (langEnv != NULL) {
 	for (i = 0; localeTable[i].lang != NULL; i++) {
 	    if (strcmp(localeTable[i].lang, langEnv) == 0) {
 		encoding = localeTable[i].encoding;
+		break;
+	    }
+	}
+	/*
+	 * There was no mapping in the locale table.  If there is an
+	 * encoding subfield, we can try to guess from that.
+	 */
+
+	if (encoding == NULL) {
+	    char *p;
+	    for (p = langEnv; *p != '\0'; p++) {
+		if (*p == '.') {
+		    p++;
+		    break;
+		}
+	    }
+	    if (*p != '\0') {
+		Tcl_DString ds;
+		Tcl_DStringInit(&ds);
+		Tcl_DStringAppend(&ds, p, -1);
+
+		encoding = Tcl_DStringValue(&ds);
+		Tcl_UtfToLower(Tcl_DStringValue(&ds));
+		if (Tcl_SetSystemEncoding(NULL, encoding) == TCL_OK) {
+		    Tcl_DStringFree(&ds);
+		    goto resetPath;
+		}
+		Tcl_DStringFree(&ds);
+		encoding = NULL;
 	    }
 	}
     }
+    if (encoding == NULL) {
+	encoding = "iso8859-1";
+    }
 
     Tcl_SetSystemEncoding(NULL, encoding);
 
+    /*
+     * Initialize the C library's locale subsystem.  This is required
+     * for input methods to work properly on X11. Note that we need to
+     * retore the initial "C" locale so that Tcl can parse numbers
+     * properly.  The side effect of setting the default locale should be to
+     * load any locale specific modules that are needed by X.
+     */
+ 
+    Tcl_DStringInit(&ds);
+    Tcl_DStringAppend(&ds, setlocale(LC_ALL, NULL), -1);
+    setlocale(LC_ALL, "");
+    setlocale(LC_ALL, Tcl_DStringValue(&ds));
+    Tcl_DStringFree(&ds);
+
+    /*
+     * In case the initial locale is not "C", ensure that the numeric
+     * processing is done in "C" locale regardless.  This is needed because
+     * Tcl relies on routines like strtod, but should not have locale
+     * dependent behavior.
+     */
+
+    setlocale(LC_NUMERIC, "C");
+ 
+
     /*
      * Until the system encoding was actually set, the library path was
      * actually in the native multi-byte encoding, and not really UTF-8
@@ -410,6 +478,7 @@ TclpSetInitialEncodings()
      * encoding.
      */
 
+    resetPath:
     pathPtr = TclGetLibraryPath();
     if (pathPtr != NULL) {
 	int objc;
@@ -451,7 +520,8 @@ TclpSetInitialEncodings()
  *	None.
  *
  * Side effects:
- *	Sets "tcl_library" and "tcl_platform" Tcl variables.
+ *	Sets "tclDefaultLibrary", "tcl_pkgPath", and "tcl_platform" Tcl
+ *	variables.
  *
  *----------------------------------------------------------------------
  */
@@ -467,7 +537,7 @@ TclpSetVariables(interp)
     char *user;
     Tcl_DString ds;
 
-    Tcl_SetVar(interp, "tcl_library", defaultLibraryDir, TCL_GLOBAL_ONLY);
+    Tcl_SetVar(interp, "tclDefaultLibrary", defaultLibraryDir, TCL_GLOBAL_ONLY);
     Tcl_SetVar(interp, "tcl_pkgPath", pkgPath, TCL_GLOBAL_ONLY);
     Tcl_SetVar2(interp, "tcl_platform", "platform", "unix", TCL_GLOBAL_ONLY);
     unameOK = 0;
